(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPairImpl imports PolyOrder Matrix
begin

datatype arctic = MinInfty | Num_arc int

datatype 'f redpair = Nat_carrier "('f,nat)lpoly_interL" 
  | Int_carrier "('f,int)lpoly_interL" 
  | Arctic_carrier "('f,arctic)lpoly_interL" 
  | Int_mat_carrier nat "('f,int mat)lpoly_interL"
  | Arc_mat_carrier nat "('f,arctic mat)lpoly_interL"

fun plus_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> arctic"
where "plus_arctic MinInfty y = y"
   |  "plus_arctic x MinInfty = x"
   |  "plus_arctic (Num_arc x) (Num_arc y) = (Num_arc (max x y))"

fun mult_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> arctic"
where "mult_arctic MinInfty y = MinInfty"
   |  "mult_arctic x MinInfty = MinInfty"
   |  "mult_arctic (Num_arc x) (Num_arc y) = (Num_arc (x + y))"

fun geq_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> bool"
where "geq_arctic x MinInfty = True"
   |  "geq_arctic MinInfty (Num_arc _) = False"
   |  "geq_arctic (Num_arc x) (Num_arc y) = (x \<ge> y)"

fun gt_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> bool"
where "gt_arctic _ MinInfty = True"
   |  "gt_arctic MinInfty (Num_arc _) = False"
   |  "gt_arctic (Num_arc x) (Num_arc y) = (x > y)"

instantiation arctic :: sshow
begin
fun sshow_arctic :: "arctic \<Rightarrow> sstring"
where "sshow_arctic (Num_arc x) = sshow x"
   |  "sshow_arctic MinInfty = ''-inf''"
instance ..
end

definition nat_name :: "sstring" where "nat_name = ''naturals''"
definition int_name :: "sstring" where "int_name = ''naturals with negative constants''"
definition int_mat_name :: "sstring" where "int_mat_name = ''matrices over the naturals with negative constants''"
definition arc_mat_name :: "sstring" where "arc_mat_name = ''matrices over the arctic semi-ring''"
definition arctic_name :: "sstring" where "arctic_name = ''arctic semiring''"


lemma int_SN : "SN {(x :: int,y). 0 \<le> y \<and> x > y}"
proof -
  have "{(x,y). 0 \<le> x \<and> (x :: int) < y} = ({(x,y). y \<ge> 0 \<and> (x :: int) > y}^-1)" by auto
  with wf_int_ge_less_than[where d = 0] have wf: "wf ({(x,y). y \<ge> 0 \<and> (x :: int) > y}^-1)" unfolding int_ge_less_than_def 
    by simp
  show "SN {(x :: int,y). 0 \<le> y  \<and> x > y}"
    by (rule wf_imp_SN, rule wf)
qed

interpretation nat_poly_carrier: poly_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True
  by (unfold_locales, auto simp: nat_distrib)


interpretation int_poly_carrier: poly_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True
  by (unfold_locales, auto simp: int_distrib Ring_and_Field.mult_right_mono Ring_and_Field.mult_left_mono)

interpretation arctic_poly_carrier: poly_carrier "\<lambda> x. True" MinInfty plus_arctic mult_arctic geq_arctic gt_arctic False
proof (unfold_locales, auto)
  fix x y
  show "plus_arctic x y = plus_arctic y x"
    by (cases x, cases y, auto, cases y, auto)
next
  fix x y z
  show "plus_arctic x (plus_arctic y z) = plus_arctic (plus_arctic x y) z"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x y z
  show "mult_arctic x (mult_arctic y z) = mult_arctic (mult_arctic x y) z"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x 
  show "mult_arctic x MinInfty = MinInfty"
    by (cases x, auto)
next
  fix x y z
  show "mult_arctic x (plus_arctic y z) = plus_arctic (mult_arctic x y) (mult_arctic x z)"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x y z
  show "mult_arctic (plus_arctic x y) z = plus_arctic (mult_arctic x z) (mult_arctic y z)"
    by (cases x, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic x y" and "gt_arctic y z"
  thus "gt_arctic x z"
    by (cases z, simp, cases y, simp, cases x, auto)
next 
  fix x y z
  assume "gt_arctic x y" and "geq_arctic y z"
  thus "gt_arctic x z"
    by (cases z, simp, cases y, simp, cases x, auto)
next
  fix x y z
  assume "geq_arctic x y" 
  thus "geq_arctic (plus_arctic x z) (plus_arctic y z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic x y" 
  thus "geq_arctic (mult_arctic x z) (mult_arctic y z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic y z" 
  thus "geq_arctic (mult_arctic x y) (mult_arctic x z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x
  show "geq_arctic x x" 
    by (cases x, auto)
next
  fix x y z
  assume "geq_arctic x y" and "geq_arctic y z"
  thus "geq_arctic x z"
    by (cases x, cases y, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z u
  assume "gt_arctic x y" and "gt_arctic z u"
  thus "gt_arctic (plus_arctic x z) (plus_arctic y u)"
    by (cases y, cases u, simp, cases z, auto, cases x, auto, cases u, auto, cases z, auto, cases x, auto, cases x, auto, cases z, auto, cases x, auto)
next
  fix x y z
  assume "gt_arctic x y"
  thus "gt_arctic (mult_arctic x z) (mult_arctic y z)"
    by (cases y, simp, cases z, simp, cases x, auto)
next
  fix x
  assume "gt_arctic MinInfty x"
  thus "x = MinInfty"
    by (cases x, auto)
qed


interpretation nat_poly_order_carrier : poly_order_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True 1 id nat_name "\<lambda> x. True"
proof (unfold_locales, auto)
  have "{(x,y). (x :: nat) < y} = ({(x,y). (x :: nat) > y}^-1)" by auto
  with wf_less have wf: "wf ({(x,y). (x :: nat) > y}^-1)" by simp 
  show "SN {(x :: nat,y). x > y}" 
    by (rule wf_imp_SN, rule wf)
qed

context poly_order_carrier
begin
declare arc_pos_max0[simp del]
end

interpretation int_poly_order_carrier : poly_order_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True 1 "max 0" int_name "\<lambda> x. True"
  by (unfold_locales, auto, rule int_SN)


fun get_arctic_num :: "arctic \<Rightarrow> int"
where "get_arctic_num (Num_arc n) = n"

fun pos_arctic :: "arctic \<Rightarrow> bool"
where "pos_arctic MinInfty = False"
    | "pos_arctic (Num_arc n) = (n \<ge> 0)"

interpretation arctic_poly_order_carrier : poly_order_carrier "\<lambda> x. True" MinInfty plus_arctic mult_arctic geq_arctic gt_arctic False "Num_arc 0" id arctic_name pos_arctic 
proof (unfold_locales, auto)
  fix x 
  show "mult_arctic (Num_arc 0) x = x" by (cases x, auto)
next
  fix x 
  show "mult_arctic x (Num_arc 0) = x" by (cases x, auto)
next
  show "SN {(x,y). pos_arctic y \<and> gt_arctic x y}" 
    unfolding SN_ars_def SN_elt_def
  proof (rule allI, rule notI)
    fix x
    assume "\<exists> f . f 0 = x \<and> (\<forall> i. (f i, f (Suc i)) \<in> {(x,y). pos_arctic y \<and> gt_arctic x y})"
    from this obtain f where "f 0 = x" and seq: "\<forall> i. (f i, f (Suc i)) \<in> {(x,y). pos_arctic y \<and> gt_arctic x y}" by auto
    from seq have steps: "\<forall> i. gt_arctic (f i) (f (Suc i)) \<and> pos_arctic (f (Suc i)) " by auto
    let ?g = "\<lambda> i. get_arctic_num (f i)"
    have "\<forall> i. ?g (Suc i) \<ge> 0 \<and> ?g i  > ?g (Suc i)"
    proof
      fix i
      from steps have i: "gt_arctic (f i) (f (Suc i)) \<and> pos_arctic (f (Suc i))" by auto
      from i obtain n where fi: "f i = Num_arc n" by (cases "f (Suc i)", simp, cases "f i", auto)
      from i obtain m where fsi: "f (Suc i) = Num_arc m" by (cases "f (Suc i)", auto)
      with i have gz: "m \<ge> 0" by simp
      from i fi fsi have "n > m" by auto
      with fi fsi gz
      show "?g (Suc i) \<ge> 0 \<and> ?g i > ?g (Suc i)" by auto
    qed
    from this obtain g where "\<forall> i. g (Suc i) \<ge> 0 \<and> (op > :: int \<Rightarrow> int \<Rightarrow> bool) (g i) (g (Suc i))" by auto
    hence "\<exists> f. f 0 = g 0 \<and> (\<forall> i. (f i, f (Suc i)) \<in> {(x,y). y \<ge> 0 \<and> x > y})" by auto
    with int_SN show False unfolding SN_ars_def SN_elt_def by blast 
  qed 
next
  fix x y
  assume "pos_arctic x" 
  thus "pos_arctic (plus_arctic x y)" by (cases x, simp, cases y, auto)
next
  fix x y
  assume "pos_arctic x" and "pos_arctic y"
  thus "pos_arctic (mult_arctic x y)" by (cases x, simp, cases y, auto)
qed

interpretation nat_mono_poly_order_carrier : mono_poly_order_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True 1 id nat_name "\<lambda> x. True" "\<lambda> x. x \<noteq> 0"
  by (unfold_locales, auto)

interpretation int_mono_poly_order_carrier : mono_poly_order_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True 1 "max 0" int_name "\<lambda> x. True" "\<lambda> x. x > 0"
  by (unfold_locales, auto simp: Ring_and_Field.mult_strict_left_mono)


interpretation int_mat_order: matrix_order "Suc n" "0 :: int" 1 "op +" "op *" "op \<ge>" "op >" "max 0" True "\<lambda> x :: int. True" 
  by (unfold_locales, auto simp: int_distrib Ring_and_Field.mult_right_mono Ring_and_Field.mult_left_mono, rule int_SN)

interpretation arc_mat_order: matrix_order "Suc n" MinInfty "Num_arc 0" plus_arctic mult_arctic geq_arctic gt_arctic id False pos_arctic 
proof (unfold_locales, auto, 
    rule arctic_poly_carrier.plus_comm, simp+,
    rule arctic_poly_carrier.compat, simp+,
    rule arctic_poly_carrier.compat2, simp+,
    rule arctic_poly_carrier.plus_left_mono, simp+,
    rule arctic_poly_carrier.times_left_mono, simp+,
    rule arctic_poly_carrier.times_right_mono, simp+,
    rule arctic_poly_carrier.geq_trans, simp+,
    simp_all add: arctic_poly_carrier.zero_leastI
                  arctic_poly_carrier.zero_leastII 
                  arctic_poly_carrier.times_gt_left_mono
                  arctic_poly_carrier.plus_gt_both_mono) 
  from arctic_poly_carrier.zero_leastI arctic_poly_order_carrier.SN
  show "SN {(x, y). gt_arctic x y \<and> pos_arctic y}" by force
qed

interpretation int_mat_poly_order_carrier :  poly_order_carrier "wf_mat_rcI (Suc n) (Suc n)" "mat0I (0 :: int) (Suc n) (Suc n)" "mat_plusI (op + :: int \<Rightarrow> int \<Rightarrow> int)" "mat_mulI (0 :: int) (op +) (op *) (Suc n)" "mat_geI (op \<ge> :: int \<Rightarrow> int \<Rightarrow> bool)" "mat_gtI (op \<ge> :: int \<Rightarrow> int \<Rightarrow> bool) (op >)" True "mat1I (0 :: int) 1 (Suc n)" "mat_max0I (max (0 :: int))" int_mat_name "mat_arc_posI (\<lambda> x. True)"
  by (rule int_mat_order.mono_poly_order_carrier, auto)

interpretation arc_mat_poly_order_carrier : poly_order_carrier "wf_mat_rcI (Suc n) (Suc n)" "mat0I MinInfty (Suc n) (Suc n)" "mat_plusI plus_arctic" "mat_mulI MinInfty plus_arctic mult_arctic (Suc n)" "mat_geI geq_arctic" "mat_geI gt_arctic" False "mat1I MinInfty (Num_arc 0) (Suc n)" "mat_max0I id" arc_mat_name "mat_arc_posI pos_arctic"
  by (rule arc_mat_order.non_mono_poly_order_carrier, auto)

abbreviation nat_components :: "nat poly_components"
where "nat_components \<equiv> ((\<lambda> x. True) , 0 , 1 , op +, op *, id, op \<ge>, op >, True, nat_name, \<lambda> x. True)"

abbreviation int_components :: "int poly_components"
where "int_components \<equiv> ((\<lambda> x. True) , 0 , 1 , op +, op *, max 0, op \<ge>, op >, True, int_name, \<lambda> x. True)"

abbreviation int_mat_components :: "nat \<Rightarrow> int mat poly_components"
where "int_mat_components n \<equiv> (wf_mat_rcI (Suc n) (Suc n), mat0I 0 (Suc n) (Suc n) , mat1I 0 1 (Suc n), mat_plusI (op +) , mat_mulI 0 (op +) (op *) (Suc n), mat_max0I (max 0), mat_geI (op \<ge>), mat_gtI (op \<ge>) (op >), True, int_mat_name, mat_arc_posI (\<lambda> x. True))"

abbreviation arctic_components :: "arctic poly_components"
where "arctic_components \<equiv> ((\<lambda> x. True), MinInfty, Num_arc 0, plus_arctic, mult_arctic, id, geq_arctic, gt_arctic, False, arctic_name, pos_arctic)"

abbreviation arc_mat_components :: "nat \<Rightarrow> arctic mat poly_components"
where "arc_mat_components n \<equiv> (wf_mat_rcI (Suc n) (Suc n), mat0I MinInfty (Suc n) (Suc n) , mat1I MinInfty (Num_arc 0) (Suc n), mat_plusI plus_arctic, mat_mulI MinInfty plus_arctic mult_arctic (Suc n), mat_max0I id, mat_geI geq_arctic, mat_geI gt_arctic, False, arc_mat_name, mat_arc_posI pos_arctic)"


fun create_rule_removal_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow>
        sstring list monad"
where "create_rule_removal_proc (Int_carrier I) = create_rule_removal_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_rule_removal_proc (Nat_carrier I) = create_rule_removal_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_rule_removal_proc (Arctic_carrier I) = (\<lambda> _ _. error [''arctic not monotonic for TRSs''])"
  |   "create_rule_removal_proc (Arc_mat_carrier n I) = (\<lambda> _ _. error [''arctic not monotonic for TRSs''])"
  |   "create_rule_removal_proc (Int_mat_carrier n I) = (\<lambda> _ _. error [''monotonic matrix-orders not yet supported''])"

fun create_redpair_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_proc (Int_carrier I) = create_redpair_proc_comp int_components I"
  |   "create_redpair_proc (Nat_carrier I) = create_redpair_proc_comp nat_components I"
  |   "create_redpair_proc (Arctic_carrier I) = create_redpair_proc_comp arctic_components I"
  |   "create_redpair_proc (Int_mat_carrier n I) = create_redpair_proc_comp (int_mat_components n) I"
  |   "create_redpair_proc (Arc_mat_carrier n I) = create_redpair_proc_comp (arc_mat_components n) I"


fun create_mono_redpair_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_mono_redpair_proc (Int_carrier I) = create_mono_redpair_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_mono_redpair_proc (Nat_carrier I) = create_mono_redpair_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_mono_redpair_proc (Arctic_carrier I) = (\<lambda> _ _ _ _. error [''arctic not monotonic for TRSs''])"
  |   "create_mono_redpair_proc (Arc_mat_carrier n I) = (\<lambda> _ _ _ _. error [''arctic not monotonic for TRSs''])"
  |   "create_mono_redpair_proc (Int_mat_carrier n I) = (\<lambda> _ _ _ _. error [''monotonic matrix-orders not yet supported''])"


fun create_redpair_ur_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_ur_proc (Int_carrier I) = create_redpair_ur_proc_comp int_components I"
  |   "create_redpair_ur_proc (Nat_carrier I) = create_redpair_ur_proc_comp nat_components I"
  |   "create_redpair_ur_proc (Arctic_carrier I) = create_redpair_ur_proc_comp arctic_components I"
  |   "create_redpair_ur_proc (Int_mat_carrier n I) = create_redpair_ur_proc_comp (int_mat_components n) I"
  |   "create_redpair_ur_proc (Arc_mat_carrier n I) = create_redpair_ur_proc_comp (arc_mat_components n) I"

fun create_mono_redpair_ur_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_mono_redpair_ur_proc (Int_carrier I) = create_mono_redpair_ur_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_mono_redpair_ur_proc (Nat_carrier I) = create_mono_redpair_ur_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_mono_redpair_ur_proc (Arctic_carrier I) = (\<lambda> _ _ _ _ _ _. error [''arctic not monotonic for TRSs''])"
  |   "create_mono_redpair_ur_proc (Arc_mat_carrier n I) = (\<lambda> _ _ _ _ _ _. error [''arctic not monotonic for TRSs''])"
  |   "create_mono_redpair_ur_proc (Int_mat_carrier n I) = (\<lambda> _ _ _ _ _ _. error [''monotonic matrix-orders not yet supported''])"

fun create_redpair_ur_af_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_ur_af_proc (Int_carrier I) = create_redpair_ur_af_proc_comp int_components I"
  |   "create_redpair_ur_af_proc (Nat_carrier I) = create_redpair_ur_af_proc_comp nat_components I"
  |   "create_redpair_ur_af_proc (Arctic_carrier I) = create_redpair_ur_af_proc_comp arctic_components I"
  |   "create_redpair_ur_af_proc (Int_mat_carrier n I) = create_redpair_ur_af_proc_comp (int_mat_components n) I"
  |   "create_redpair_ur_af_proc (Arc_mat_carrier n I) = create_redpair_ur_af_proc_comp (arc_mat_components n) I"


lemma create_rule_removal_proc: 
  shows "ci_tt_impl (\<lambda> R rR. isOK(create_rule_removal_proc I R rR))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI R rR))" by auto
  show ?thesis by (simp only: id, rule nat_mono_poly_order_carrier.rule_removal_tt_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp int_components (\<lambda> x. x > 0) nI R rR))" by auto
  show ?thesis by (simp only: id, rule int_mono_poly_order_carrier.rule_removal_tt_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding ci_tt_impl_def by auto
next
  case (Int_mat_carrier n nI) thus ?thesis unfolding ci_tt_impl_def by auto
next
  case (Arc_mat_carrier n nI) thus ?thesis unfolding ci_tt_impl_def by auto
qed


lemma create_redpair_proc: "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp nat_components nI P rP R))" by auto
  show ?thesis by (simp only: id, rule nat_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp int_components nI P rP R))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp arctic_components nI P rP R))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Int_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (int_mat_components n) nI P rP R))" by auto
  show ?thesis by (simp only: id, rule int_mat_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (arc_mat_components n) nI P rP R))" by auto
  show ?thesis by (simp only: id, rule arc_mat_poly_order_carrier.redpair_proc_comp_ci)
qed


lemma create_mono_redpair_proc: "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI P rP R rR))" by auto
  show ?thesis by (simp only: id, rule nat_mono_poly_order_carrier.mono_redpair_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp int_components (\<lambda> x. x > 0) nI P rP R rR))" by auto
  show ?thesis by (simp only: id, rule int_mono_poly_order_carrier.mono_redpair_proc_comp_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Int_mat_carrier n nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Arc_mat_carrier n nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
qed


lemma create_redpair_ur_proc: "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp nat_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule nat_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp int_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp arctic_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Int_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (int_mat_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_mat_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (arc_mat_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arc_mat_poly_order_carrier.redpair_ur_proc_comp_ci)
qed


lemma create_redpair_ur_af_proc: "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp nat_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule nat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next 
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp int_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp arctic_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Int_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (int_mat_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_mat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (arc_mat_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arc_mat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
qed


lemma create_mono_redpair_ur_proc: "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id, rule nat_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp int_components (\<lambda> x. x > 0) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id, rule int_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Int_mat_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Arc_mat_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
qed

  
end